/* 
 * Copyright (C) 2020 - 2024, Chengdu BeiZhongWangXin Technology Co., Ltd. All rights reserved.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <fcntl.h>
#include <linux/sockios.h>
#include <linux/netlink.h>
#include <linux/ethtool.h>
#include "crc32.h"

typedef unsigned char U8;
const char* PkgSIg = "BZWXNE6SNICFWPKG";

#define IMAGE_PATH			"/usr/lib/firmware/"
#define IMAGE_COUNT			(ImgMax - ImgLoader)

typedef enum ImgType
{
	ImgLoader = 0xB0,
	ImgApp,
	ImgNp,
	ImgPxe,
    ImgCfg,
    ImgMax,
} ImgType_t;

typedef struct FWInfo
{
	U8 type;
	U8 rsvd[7];
	uint32_t offset;
	uint32_t length;
} FWInfo_t;

typedef struct pkgHeader
{
	char flags[16];

	uint32_t hdrVer;
	uint32_t hdrLen;
	uint32_t pkgVer;
	uint32_t pkgLen;

	uint32_t buildDate;
	uint32_t buildTime;
	uint32_t secCnt;
	uint32_t secOffset;

	U8 cipher;
	U8 rsvd[7];
	uint32_t pkgCRC;
	uint32_t hdrCRC;

	FWInfo_t Infos[0];
} pkgHeader_t;

typedef struct upgrade
{
	char* appName;
	char infName[64];
	char filename[256];

	bool hasInf;
	bool hasFile;

    bool isUpdateAll;
    bool toUpgrade[IMAGE_COUNT];

	uint8_t* fileData;
	uint32_t fileLen;
	uint32_t sectCount;

	char** setFileNames;

	int32_t sock;
	struct ifreq ifr;
} upgrade_t;

static upgrade_t upgrade = {
	.infName = { 0 },
	.filename = { 0 },

	.hasInf = false,
	.hasFile = false,

    .isUpdateAll = true,
    .toUpgrade = {false},
};

static char* GetFilename(ImgType_t type)
{
	char* name = "misc.bin";
	switch (type)
	{
	case ImgLoader:
		name = "Loader.bin";
		break;

	case ImgApp:
		name = "App.bin";
		break;

	case ImgNp:
		name = "NP.bin";
		break;

	case ImgPxe:
		name = "NceSnpDriver.bin";
		break;
    
    case ImgCfg:
		name = "CFG.bin";
		break;

	default:
		// do nothing
		break;
	}

	return name;
}

static bool Isvalidsignature(const int8_t *signature, const char *expectation)
{
    int8_t ch;
    bool isValid = true;

    while ('\0' != (ch = *expectation++)) {
        if (ch != *signature++) {
            isValid = false;
            break;
        }
    }

    return isValid;
}

static bool IsHeaderIntact(pkgHeader_t *const header)
{
	if (!Isvalidsignature(header->flags, PkgSIg)) {
		return false;
	}
    uint32_t expectation = header->hdrCRC;
    header->hdrCRC = DUMMY_CRC;
    uint32_t crc = Crc32(header, header->hdrLen);
    header->hdrCRC = expectation;
    return (expectation == crc);
}

static bool IsValidImage(void *image)
{
    pkgHeader_t *hdr = (pkgHeader_t *)image;
    uint32_t expectation = hdr->pkgCRC;
    uint32_t tmp = hdr->hdrCRC;
    hdr->hdrCRC = DUMMY_CRC;
    hdr->pkgCRC = DUMMY_CRC;
    uint32_t crc = Crc32(image, hdr->pkgLen);
    hdr->pkgCRC = expectation;
    hdr->hdrCRC = tmp;
    return (expectation == crc);
}

static bool CheckValid(void *image)
{
	if (!IsHeaderIntact(image)) {
		printf("bad fw: header error\n");
		return false;
	}

	if (!IsValidImage(image)) {
		printf("bad fw: pkt error\n");
		return false;
	}

	return true;
}

static void usage(char* appName)
{
	printf(
		"\r\n"
		"Usage: %s [-h] [-l] [-a] [-n] [-p] -i <interface> -f <package> \r\n"
		" -h             Print this message and exit\r\n"
		" -i [device]    Interface to use, for example 'eth0'\r\n"
		" -f [filename]  Specify the FW package filename, for example 'FW-nnn.bin'\r\n"
		" -l             Update Loader\r\n"
		" -a             Update App\r\n"
		" -n             Update NP\r\n"
		" -p             Update PXE\r\n"
        " -c             Update CFG\r\n\r\n"
		" If none of l/a/n/p/c specified, then all parts will be updated.\r\n", appName
		);
}

static int32_t upgrade_parse_ops(upgrade_t *upg, uint32_t argc, char* argv[])
{
	uint32_t c;
	char* appName = argv[0];

	if(argc < 2)
	{
	    usage(appName);
	    return -1;
	}

	upg->appName = appName;
	upg->isUpdateAll = true;
	upg->setFileNames = NULL;
	upg->fileData = NULL;

	while (EOF != (c = getopt(argc, argv, "lanphi:f:")))
	{
		switch (c)
		{
		case 'l':
            upg->toUpgrade[ImgLoader - ImgLoader] = true;
            upg->isUpdateAll = false;
			break;
		case 'a':
            upg->toUpgrade[ImgApp - ImgLoader] = true;
            upg->isUpdateAll = false;
			break;
		case 'n':
            upg->toUpgrade[ImgNp - ImgLoader] = true;
            upg->isUpdateAll = false;
			break;
		case 'p':
            upg->toUpgrade[ImgPxe - ImgLoader] = true;
            upg->isUpdateAll = false;
			break;
        case 'c':
            upg->toUpgrade[ImgCfg - ImgLoader] = true;
            upg->isUpdateAll = false;
			break;
		case 'i':
			sprintf(upg->infName, "%s", optarg);
			upg->hasInf = true;
			break;
		case 'f':
			sprintf(upg->filename, "%s", optarg);
			upg->hasFile = true;
			break;
		case 'h':
			usage(appName);
			upg->appName = appName;
			return 0;
		default:
			printf("Unsupported argument!\n");
			return -1;
		}
	}

	if (!(upg->hasInf && upg->hasFile))
	{
		printf("Please provide interface through -i, provide file through -f\n");
		return -1;
	}

    if (upg->isUpdateAll) {
        for (uint32_t i = 0; i < IMAGE_COUNT; i++) {
             upg->toUpgrade[i] = true;
        }
    }

	return 0;
}


static int32_t upgrade_read_package(upgrade_t *upg)
{
	FILE* infile = fopen(upg->filename, "r");
	if (NULL == infile)
	{
		printf("Open %s failed\n", upg->filename);
		return -1;
	}
	
    fseek(infile, 0, SEEK_END);
    upg->fileLen = ftell(infile);

	upg->fileData = (uint8_t*)malloc(sizeof(uint8_t) * upg->fileLen);
	if (upg->fileData == NULL) {
		fclose(infile);
		printf("%s malloc error\n", __FUNCTION__);
		return -1;
	}

	fseek(infile, 0, SEEK_SET);
	fread(upg->fileData, 1, upg->fileLen, infile);
	fclose(infile);

	return 0;
}

static int32_t upgrade_check_package(upgrade_t *upg) 
{
	pkgHeader_t *header = (pkgHeader_t *)upg->fileData;
	if (header->pkgLen != upg->fileLen) {
		printf("bad fw: total len:%d, len in pkt:%d\n", upg->fileLen, header->pkgLen);
		return -1;
	}

	Crc32Init();

	if(!CheckValid(upg->fileData)) {
		return -1;
	}

	upg->sectCount = header->secCnt;
    if (upg->sectCount < 0 || upg->sectCount > (ImgMax-ImgLoader)) {
        printf("bad fw: error secConut:%d\n", upg->sectCount);
        return -1;
    }

	return 0;
}

static int32_t upgrade_create_temp_file(upgrade_t *upg)
{	
	uint32_t sectCount = upg->sectCount;
	uint8_t *tmpFw =upg->fileData;
	char** fwNames = NULL;

	fwNames = malloc(sizeof(char *) * sectCount);
	if (fwNames == NULL) {
		return -1;
	}
	
	upg->setFileNames = fwNames;
	FWInfo_t infos[sectCount];
	memcpy(infos, (const void *)&tmpFw[sizeof(pkgHeader_t)], sectCount * sizeof(FWInfo_t));

    for(uint32_t i = 0; i < sectCount; i++) {
        ImgType_t type = (ImgType_t)infos[i].type;
        if (type < ImgLoader || type >= ImgMax) {
            printf("bad fw: error secType:0x%08x\n", type);
            return -1;
        }

        if (upg->toUpgrade[type - ImgLoader]) {
            uint32_t offset = infos[i].offset;
            uint32_t length = infos[i].length;

            char* name = GetFilename(type);
            fwNames[i] = name;
            uint32_t nameLen = strlen(IMAGE_PATH) + strlen(name);
            char* fullPath = malloc(nameLen + 16);

            strcpy(fullPath, IMAGE_PATH);
            strcat(fullPath, name);

            U8 tmp[length];
			memcpy(tmp, (const void *)&tmpFw[offset], length);

            FILE* outFile = fopen(fullPath, "w");
			free(fullPath);

            if(NULL != outFile) {
                fwrite(tmp, 1, length, outFile);
                fclose(outFile);
            }
            else {
                printf("Error: Open file %s for write failed!", name);
                return -1;
            }
        }
        else {
            fwNames[i] = NULL;
        }
    }

	return 0;
}

static int32_t upgrade_open_socket(upgrade_t *upg)
{
	const char* ifname = upg->infName;
	struct ifreq *ifr = &upg->ifr;

	if (strlen(ifname) >= IFNAMSIZ) {
		printf("Device name longer than characters\n");
	}

	strcpy(ifr->ifr_name, ifname);

	/* Open control socket. */
	upg->sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (upg->sock < 0)
		upg->sock = socket(AF_NETLINK, SOCK_RAW, NETLINK_GENERIC);
	if (upg->sock < 0) {
		perror("Cannot get control socket");
		return -1;
	}

	if (ioctl(upg->sock, SIOCGIFINDEX, ifr) == -1) {
		return -1;
	}

	return 0;
}

static int32_t upgrade_by_ethtool(upgrade_t *upg)
{
	int32_t ret = 0;
	char** fwNames = upg->setFileNames;
	uint32_t sectCount = upg->sectCount;
	struct ifreq *ifr = &upg->ifr;
	struct ethtool_flash flash_cmd;
	char part[IMAGE_COUNT + 1] = {'\0'};
	
	printf("%s ","Start Upgrade");
	for(uint32_t i = 0; i < sectCount; i++) {
        if (fwNames[i] == NULL) {
            continue;
        }

		memset(&flash_cmd, 0, sizeof(flash_cmd));
		flash_cmd.cmd = ETHTOOL_FLASHDEV;
		strcpy(flash_cmd.data, fwNames[i]);	
		ifr->ifr_data = (char *)&flash_cmd;
		ret = ioctl(upg->sock, SIOCETHTOOL, ifr);
		if (ret) {
			break;
		}
		part[i] = '.';
		printf("\r%s%s","Start Upgrade",part);
		fflush(stdout);
        sleep(1);
	}
	
	printf("\n");
	return ret;
}

static int32_t upgrade_firmware(upgrade_t *upg)
{
	int32_t ret = 0;

	if (0 != upgrade_open_socket(upg)) {
		ret = -1;
		goto _close_sock;
	}

    ret = upgrade_by_ethtool(upg); 
	if (ret) {
		printf("Upgrade failed: %d\n", ret);
	}
	
_close_sock:
	if (upg->sock >= 0) {
		close(upg->sock);
	}
	return ret;
}

static int32_t upgrade_del_temp_file(upgrade_t *upg)
{
	char** fwNames = upg->setFileNames;
	uint32_t sectCount = upg->sectCount;

	for(uint32_t i = 0; i < sectCount; i++) {
		if (fwNames[i] == NULL) {
            continue;
        }
		char* tmpfullPath = malloc(strlen(IMAGE_PATH) + strlen(fwNames[i]) + 16);
		strcpy(tmpfullPath, IMAGE_PATH);
        strcat(tmpfullPath, fwNames[i]);
		remove(tmpfullPath);
	}

	return 0;
}

static void upgrade_clr(upgrade_t *upg)
{
	if (NULL != upg->fileData) {
		free(upg->fileData);
	}

	if (NULL != upg->setFileNames) {
		free(upg->setFileNames);
	}

}

int32_t main(uint32_t argc, char* argv[])
{
	int32_t ret = 0;
	upgrade_t *upg = &upgrade;
	
	if (0 != upgrade_parse_ops(upg, argc, argv)) {
		return -1;
	}

	if(0 != upgrade_read_package(upg)) {
		ret = -1;
		goto _exit;
	}

	if(0 != upgrade_check_package(upg)) {
		ret = -1;
		goto _exit;
	}

	if(0 != upgrade_create_temp_file(upg)) {
		ret = -1;
		goto _exit;
	}

	if(0 != upgrade_firmware(upg)) {
		ret = -1;
		goto _exit;
	}
	upgrade_del_temp_file(upg);
	
	printf("Upgrade success\n");

_exit:
	upgrade_clr(upg);

	return ret;
}
